from quart import Quart, request, render_template, redirect
from math import ceil
import peertube
import sys

# Import everything from utils.py, to keep main.py focused
# on CLI arguments parsing, and web routes declarations
from utils import *

cache = peertube.Cache()
api = peertube.API(cache)
# Let's follow subscriptions from the current working dir
subscriptions = Subscriptions('.', api, cache)

# --- WEBSERVER STUFF ---
# All utilities are above this point.
# Routes are defined below
app = Quart(__name__)

# Class to build rendering context from, can be chained like
# context = Context().paginate(URL, start=0, count=10)
#                    .instance(domain)
# TODO: Since some context parsing will in fact produce fallible requests, how to error?
# Do we store self.failures(default None)? Then render can check for failures and render error.html instead
class Context():
    def __init__(self, api):
        self.context = {}
        self.failures = []
        self.api = api

    # Helper function that's called by the lambda for passing context
    def build(self, _ignored_old_api):
        return self.context

    # Helper function that's called during render for passing errors
    def failed(self):
        return self.failures

    # Helper function to insert a successful result into context, or register failure otherwise
    # TODO: Don't actually query the API when there was already a failure, so we can chain calls worryfree
    def insert_result(self, key, result):
        if isinstance(result, FailedRequest):
            self.failures.append(result)
        else:
            self.context[key] = result
        return self

    # Helper function to insert a raw value into context. Will not insert FailedRequests, and it should
    # not be provided any in all cases.
    def insert(self, key, value):
        if isinstance(value, FailedRequest):
            print("[ERROR] Context.insert() called with a failed request:\n%s", value)
            return self
        self.context[key] = value
        return self

    # Build pagination context. Since total page count is fetched from 
    def paginate(self, url, page, total_pages):
        # TODO: Actually i think pagination URL takes no page number, as it's set in templates?
        self.context["paginate_url"] = url.replace("${page}", str(page))
        self.context["page"] = page
        self.context["pages_total"] = total_pages
        return self

    # Build pagination context from API results, that were previously inserted
    # with self.insert_result(), so you don't have to store the API result in the route
    # declaration, if you don't need to process it further
    # DON'T CALL ME FROM INDEX ROUTE, SEE NOTE BELOW
    def paginate_results(self, results_key, url, page):
        if results_key not in self.context:
            self.failures.append(FailedRequest("%s not found in pagination context. Either a previous request for results to paginate has failed, or this is a programming error worth reporting." % results_key))
            return self
        # TODO: Currently there is no handling of list of results, such as on the subscription page
        # because the only page using it is index, and index currently doesn't support pagination
        # because the merging algorithm is not smart
        results = self.context[results_key]
        if isinstance(results, FailedRequest):
            # Request failed so we don't build the pagination
            # Also this branch should never be reached because self.insert_result() should prevent it
            print("[ERROR] Tried to paginate a failed result. You should insert results with context.insert_result, not context.insert to prevent this mistake.")
            return self
        # TODO: Pagination should not be done from the returned number of results,
        # (because that's only `count` items), but from result["total"] which should be kept until here
        return self.paginate(url, page, ceil(len(results) / 10))

    # Build instance context (eg. domain and instance_name)
    def instance(self, domain):
         self.context["domain"] = domain
         self.insert_result("instance", api.instance_name(domain))
         return self

    # Build video context (captions, quality, embed, comments) from domain, id and request headers
    def video(self, domain, id, args={}):
        video = self.api.video(domain, id)
        self.insert_result("video", video)
        if isinstance(video, FailedRequest):
            return self
        
        self.insert_result("video_info", VideoInfo(api, video, args)) \
            .insert_result("captions", api.video_captions(domain, id)) \
            .insert_result("comments", format_comments(api.video_comments(domain, id)) if video["commentsEnabled"] else []) \
            .insert("quality", args.get("quality")) \
            .insert("embed", args.get("embed"))
        return self

# --- INDEX ROUTE ---
@app.route("/")
async def main():
    info, info_failures = subscriptions.info()
    videos, videos_failures = subscriptions.videos(limit=12)
    # Inside subscriptions variable, you may find either an account info structure, or a channel info structure. Channels may be recognized due to `ownerAccount` property.

    return await render_template(
        "index.html",
        videos=videos,
        subscriptions=info,
        videos_failures=format_errors_html(videos_failures),
        info_failures=format_errors_html(info_failures),
    )
# --- END INDEX ROUTE ---

# --- SEARCH ROUTES ---
@app.route("/search", methods = ["POST"])
async def search_post_redirect():
    query = (await request.form)["query"]
    return redirect("/search/" + query)

@app.route("/search", methods = ["GET"])
async def search_get_redirect():
    query = request.args.get("query")
    return redirect("/search/" + query)

@app.route("/<string:domain>/search", methods=["POST"])
async def search_instance_redirect(domain):
    query = (await request.form)["query"]
    return redirect("/" + domain + "/search/" + query)

@app.route("/search/<string:query>", defaults = {"page": 1})
@app.route("/search/<string:query>/<int:page>")
async def search(query, page):
    context = Context(api).insert_result("results", api.search(query, (page - 1) * 10)) \
                          .paginate_results("results", "/search" + query + "/", page, count=10) \
                          .insert("query", query)
    return await render("simpleer_search_results.html", context.failed(), context.build)

@app.route("/<string:domain>/search/<string:query>", defaults = {"page": 1})
@app.route("/<string:domain>/search/<string:query>/<int:page>")
async def search_instance(domain, query, page):
    context = Context(api).instance(domain) \
                          .insert_result("results", api.search_instance(domain, query, (page - 1) * 10)) \
                          .paginate_results("results", "/search" + query + "/", page, count=10) \
                          .insert("search_term", query) # TODO: rename query as in cross-instance search?
    return await render("search_results.html", context.failed(), context.build)
# --- END SEARCH ROUTES ---

# --- INSTANCE ROUTES ---
@app.route("/<string:domain>/")
async def instance(domain):
    # favicon.ico is not a domain name
    if domain == "favicon.ico": return await favicon()
    return redirect("/" + domain + "/videos/trending")

@app.route("/<string:domain>/videos/<string:category>", defaults = {"page": 1})
@app.route("/<string:domain>/videos/<string:category>/<int:page>")
async def instance_videos(domain, category, page):
    if category not in [ "most-liked", "local", "trending", "recently-added" ]:
        return await render_custom_error("No such video category: %s" % category, 404)

    context = Context(api).instance(domain) \
                          .insert_result("videos", api.instance_videos(domain, category=category, start=(page - 1) * 10)) \
                          .paginate_results("results", "/" + domain + "/videos/" + category + '/', page, count=10)
    
    return await render("instance/%s.html" % category, context.failed(), context.build)
# --- END INSTANCE ROUTES ---

# --- VIDEO ROUTE ---
@app.route("/<string:domain>/videos/watch/<string:id>/")
async def video(domain, id):
    context = Context(api).instance(domain) \
                          .video(domain, id, request.args)
    return await render("video.html", context.failed(), context.build)

@app.route("/<string:domain>/videos/watch/<string:id>/<string:lang>.vtt")
async def captions(domain, id, lang):
    captions = api.video_captions(domain, id)
    if isinstance(captions, FailedRequest):
        return await render_error(captions)

    for entry in captions:
        if entry["language"]["id"] == lang:
            content = api.video_captions_proxy(domain, entry["captionPath"].split('/')[-1])
            return await render_plaintext(content)

    return await render_custom_error("This video has no subtitles/captions in requested language", code=404)
# --- END VIDEO ROUTE ---

# --- ACCOUNT ROUTES ---
@app.route("/<string:domain>/accounts/<string:name>")
async def accounts_redirect(domain, name):
    return redirect("/" + domain + "/accounts/" + name + "/video-channels")

@app.route("/<string:domain>/accounts/<string:name>/video-channels", defaults = {"page": 1})
@app.route("/<string:domain>/accounts/<string:name>/video-channels/<int:page>")
async def account_channels(domain, name, page):
    # TODO: Don't do useless requests since a request already failed? \
    context = Context(api).insert_result("video_channels", api.account_channels(domain, name, start = (page - 1) * 10, count=10)) \
                          .paginate_results("video_channels", "/" + domain + "/accounts/" + name + "/video-channels/${page}", page) \
                          .insert_result("account", api.account(domain, name)) \
                          .instance(domain) \
                          .insert("name", name)
    return await render("accounts/channels.html", context.failed(), context.build)

@app.route("/<string:domain>/accounts/<string:name>/videos", defaults = {"page": 1})
@app.route("/<string:domain>/accounts/<string:name>/videos/<int:page>")
async def account_videos(domain, name, page):
    # TODO: Don't do useless requests since a request already failed?
    context = Context(api).insert_result("videos", api.account_videos(domain, name, start = (page - 1) * 10, count=10)) \
                          .paginate_results("videos", "/" + domain + "/accounts/" + name + "/videos/${page}", page) \
                          .insert_result("account", api.account(domain, name)) \
                          .instance(domain) \
                          .insert("name", name)
    return await render("accounts/videos.html", context.failed(), context.build)

@app.route("/<string:domain>/accounts/<string:name>/about")
async def account_about(domain, name):
    context = Context(api).instance(domain) \
                          .insert_result("account", api.account(domain, name)) \
                          .insert("name", name)
    return await render("accounts/about.html", context.failed(), context.build)
# --- END ACCOUNT ROUTES ---

# --- CHANNEL ROUTES ---
@app.route("/<string:domain>/video-channels/<string:name>")
async def video_channels_redirect(domain, name):
    return redirect("/" + domain + "/video-channels/" + name + "/videos")

@app.route("/<string:domain>/video-channels/<string:name>/videos", defaults = {"page": 1})
@app.route("/<string:domain>/video-channels/<string:name>/videos/<int:page>")
async def channel_videos(domain, name, page):
    channel = api.channel(domain, name)
    if isinstance(channel, FailedRequest):
        return await render_error(channel)
    # TODO: Is it good to use ownerAccount.{host,name}?
    context = Context(api).insert_result("video_channel", channel) \
                          .insert_result("videos", api.channel_videos(domain, name, start = (page - 1) * 10, count = 10)) \
                          .paginate_results("videos", "/" + domain + "/video-channels/" + name + "/videos/${page}", page) \
                          .instance(domain) \
                          .insert_result("account", api.account(channel["ownerAccount"]["host"], channel["ownerAccount"]["name"])) \
                          .insert("name", name)
    return await render("video_channels/videos.html", context.failed(), context.build)

# TODO: Hope this route works, i don't know some account who has channel playlists to test with
@app.route("/<string:domain>/video-channels/<string:name>/video-playlists", defaults = {"page": 1})
@app.route("/<string:domain>/video-channels/<string:name>/video-playlists/<int:page>")
async def channel_playlists(domain, name, page):
    channel = api.channel(domain, name)
    if isinstance(channel, FailedRequest):
        return await render_error(channel)
    # TODO: Is it good to use ownerAccount.{host,name}?
    context = Context(api).insert_result("video_playlists", api.channel_playlists(domain, name, (page - 1) * 10)) \
                          .paginate_results("video_playlists", "/" + domain + "/video-channels/"  + name + "/video-playlists/${page}", page) \
                          .insert_result("video_channel", channel) \
                          .instance(domain) \
                          .insert_result("account", api.account(channel["ownerAccount"]["host"], channel["ownerAccount"]["name"])) \
                          .insert("name", name)
    return await render("video_channels/video_playlists.html", context.failed(), context.build)

@app.route("/<string:domain>/video-channels/<string:name>/about")
async def channel_about(domain, name):
    context = Context(api).instance(domain) \
                          .insert_result("video_channel", api.channel(domain, name)) \
                          .insert("name", name)
    return await render("video_channels/about.html", context.failed(), context.build)

# --- DIVERSE ROUTES ---
@app.route("/favicon.ico")
async def favicon():
    return await render_custom_error("We don't have a favicon yet. If you would like to contribute one, please send it to ~metalune/public-inbox@lists.sr.ht", code=404)

@app.route("/opensearch.xml")
async def opensearch():
    try:
        with open('opensearch.xml', 'r') as f:
            return f.read().replace('$BASEURL', request.headers["Host"])
    except Exception as e:
        return await render_error(FailedRequest(e))
# --- END DIVERSE ROUTES ---
# --- END WEBSERVER STUFF ---

# --- CLI STUFF ---
if __name__ == "__main__":
    if len(sys.argv) == 3:
        interface = sys.argv[1]
        port = sys.argv[2]
    elif len(sys.argv) == 2:
        interface = "127.0.0.1"
        port = sys.argv[1]
    else:
        interface = "127.0.0.1"
        port = "5000"
    app.run(host=interface, port=port)
# --- END CLI STUFF ---
