from bs4 import BeautifulSoup
import requests
import json
import sys
from datetime import datetime
from utils import FailedRequest

# Helper Class for using caches, you can use any other cache that implements the same API
# Default TTL: 3600s (1h)
class Cache:
    def __init__(self, ttl=3600):
        self.dict = {}
        self.ttl = ttl

    # Use ttl=0 to disable
    def get(self, key, ttl=None):
        if ttl == None: ttl = self.ttl
        if key in self.dict:
            last_time_updated = (self.dict[key])[1]
            time_diff = datetime.now() - last_time_updated

            if time_diff.total_seconds() > ttl:
                # Outdated data
                return None
            # Data still valid according to TTL
            return (self.dict[key])[0]

        else:
            # No data
            return None

    def set(self, key, value):
        self.dict[key] = [ value, datetime.now() ]

def html_title(content):
    soup = BeautifulSoup(content, "lxml")
    title = soup.find('title')
    if title:
        return title.text
    else:
        return "PeerTube Instance"

class API:
    # The PeertubeAPI is initialized with a caching backend and a default TTL, that can be overriden in specific
    # API request calls. The caching backend should implement a get(key, ttl) and set(key, value) API.
    # Also can specify the SepiaSearch instance here
    def __init__(self, cache, ttl=3600, search="https://search.joinpeertube.org"):
        self.cache = cache
        self.ttl = ttl
        # If search instance has no protocol set, assume https://
        if not search.startswith("http"): search = "https://" + search
        # Remove trailing slash
        if search.endswith('/'): self.search_source = search[0:-1]
        else: self.search_source = search
        # Instance search API expects a filter and a sorting criteria. Here we provide
        # higher-level categories that match requested URIs: trending, most-liked, recently-added, local
        self.instance_filters = {
            "trending": (None, "-trending"),
            "most-liked": (None, "-likes"),
            "local": ("local", "-publishedAt"),
            "recently-added": (None, "-publishedAt"),
        }

    # Wrapper around requests.get() so that it cannot fail
    # Usually, you want to call self.request(), which is an even higher-level wrapper
    #  If the request succeeds:
    #    - parse as JSON, or return None
    #    - if "error" field in JSON, return None
    #    - return the parsed JSON
    #  Else: return None
    #  In all cases where the function returns None, the errors are logged
    # Use ttl=0 to disable Cache for this request
    # Only use with JSON requests, otherwise use self.(plaintext_request)
    def json_request(self, url):
        print("[DEBUG] Requesting JSON URL %s" % url)
        try:
            # If serialization fails, we'll end up in `except` block
            parsed_response = json.loads(requests.get(url).text)
            if "error" in parsed_response:
                print("[WARN] Remote peertube returned error for %s:\n%s" % (url, parsed_response["error"]))
                return FailedRequest("Remote peertube server returned an error for URL %s:\n%s" % (url, parsed_response["error"]))
            return parsed_response
        except Exception as e:
            print("[WARN] Error fetching page \"%s\":\n%s" % (url, e))
            return FailedRequest(e)

    # Wrapper around requests.get() so that it cannot fail
    # Use with plaintext requests, for JSON requests use self.request()
    def plaintext_request(self, url):
        print("[DEBUG] Requesting plaintext URL %s" % url)
        try:
            return requests.get(url).text
        except Exception as e:
            print("[WARN] Error fetching page \"%s\":\n%s" % (url, e))
            return FailedRequest(e)

    # Useful wrapper method to reduce boilerplate
    # args: parameters tuple to form the cache key
    # url: string template to form URL from args tuple, where ${i} is arg[i]
    # key: Key to extract from a successful response
    # backend: the method to use for fetching URL (default: self.json_request), can be self.plaintext_request
    # extractor: a lambda function to execute to extract stuff from a successful request, when key isn't set
    def request(self, args, url, key=None, ttl=None, backend=None, extractor=None):
        # WTF python? '/'.join(("foo)) => "f/o/o", not "foo"?! Special case when only one arg
        if isinstance(args, str): args = (args, )
        cache_key = '/'.join(args)
        cached = self.cached(cache_key, ttl=ttl)
        if cached == None:
            # Defaults to making a JSON API request
            if backend == None: backend = self.json_request
            for i in range(0, len(args)):
                url = url.replace("${%s}" % str(i), args[i])
            res = backend(url)
            if not isinstance(res, FailedRequest):
                if key != None:
                    # Extract requested key from successful request
                    res = res[key]
                elif extractor != None:
                    # Run extractor on result from successful request
                    res = extractor(res)
            self.save(cache_key, res)
            return res
        return cached

    # Returns an entry from cache, automatically prefixing "peertube-FUNC-" where FUNC is the caller method name
    def cached(self, key, ttl=None):
        if ttl == None: ttl = self.ttl
        caller = sys._getframe(2).f_code.co_name
        key_name = "peertube-" + caller + "-" + key
        res = self.cache.get(key_name, ttl=ttl)
        if res == None:
            print("[CACHE] Entry not found for %s: %s" % (caller, key))
            return None
        print("[CACHE] Found entry for %s: %s" % (caller, key))
        return res

    # Save an entry into cache, automatically prefixing "peertube-FUNC-" where FUNC is the caller method name
    def save(self, key, value):
        # TODO: Maybe now that we call from two levels above (request_wrapper->request->save) we need to adapt the frame fetched
        caller = sys._getframe(2).f_code.co_name
        key_name = "peertube-" + caller + "-" + key
        print("[CACHE] Saving entry for %s: %s" % (caller, key))
        self.cache.set(key_name, value)

    # Fetch instance name from its HTML source
    def instance_name(self, domain, ttl=None):
        return self.request(
            (domain),
            "https://${0}",
            backend=self.plaintext_request,
            extractor=html_title
        )

    # Search the configured self.search_source for `query`, returning `count` items after `start`
    def search(self, query, start=0, count=10, ttl=None):
        return self.request(
            (str(start), str(count), self.search_source, query),
            # self.search_source already has protocol pre-pended
            "${2}/api/v1/search/videos?start=${0}&count=${1}&search=${3}",
            key="data"
        )

    # Search a specific Peertube instance  for `query`,
    # returning `count` items after `start`. Slightly different URL format from SepiaSearch (self.search())
    def search_instance(self, domain, term, start=0, count=10, ttl=None):
        return self.request(
            (str(start), str(count), domain, term),
            "https://${2}/api/v1/search/videos?start=${0}&count=${1}&search=${3}&sort=-match&searchTarget=local",
            key="data"
        )

    # Default category is local, other categories are: trending, most-liked, recently-added, local
    # See self.instance_filters in self.__init()__ for sorting/filtering detail
    def instance_videos(self, domain, start=0, count=10, category="local", ttl=None):
        if not category in self.instance_filters:
            return FailedRequest("instance_videos called with bogus filter: %s" % category)
        filt, sort = self.instance_filters[category]
        url = "https://${2}/api/v1/videos?sort=" + sort + "&start=${0}&count=${1}"
        if filt: url += "&filter=" + filt
        return self.request(
            (str(start), str(count), domain, category),
            url,
            key="data"
        )

    def video(self, domain, id, ttl=None):
        return self.request(
            (domain, id),
            "https://${0}/api/v1/videos/${1}"
        )

    def video_captions(self, domain, id, ttl=None):
        return self.request(
            (domain, id),
            "https://${0}/api/v1/videos/${1}/captions",
            key="data"
        )
    
    def video_captions_proxy(self, domain, caption_id, ttl=None):
        # URL is hardcoded to prevent further proxying. URL may change with updates, see captions API
        # eg. https://kolektiva.media/api/v1/videos/9c9de5e8-0a1e-484a-b099-e80766180a6d/captions
        # TODO: What if the captionPath doesn't follow this format on an instance? Should we really proxy ANYTHING returned by API?
        return self.request(
            (domain, caption_id),
            "https://${0}/lazy-static/video-captions/${1}",
            backend=self.plaintext_request
        )

    def video_comments(self, domain, id, ttl=None):
        return self.request(
            (domain, id),
            "https://${0}/api/v1/videos/${1}/comment-threads",
            key="data"
        )

    def account(self, domain, name, ttl=None):
        return self.request(
            (domain, name),
            "https://${0}/api/v1/accounts/${1}"
        )

    def account_channels(self, domain, name, start=0, count=10, ttl=None):
        return self.request(
            (str(start), str(count), domain, name),
            "https://${2}/api/v1/accounts/${3}/video-channels?start=${0}&count=${1}",
            key="data"
        )

    def account_videos(self, domain, name, start=0, count=10, ttl=None):
        return self.request(
            (str(start), str(count), domain, name),
            "https://${2}/api/v1/accounts/${3}/videos?start=${0}&count=${1}",
            key="data"
        )

    def channel(self, domain, name, ttl=None):
        return self.request(
            (domain, name),
            "https://${0}/api/v1/video-channels/${1}"
        )

    def channel_videos(self, domain, name, start=0, count=10, ttl=None):
        return self.request(
            (str(start), str(count), domain, name),
            "https://${2}/api/v1/video-channels/${3}/videos?start=${0}&count=${1}",
            key="data"
        )

    def channel_playlists(self, domain, name, start=0, count=10, ttl=None):
        return self.request(
            (str(start), str(count), domain, name),
            "https://${2}/api/v1/video-channels/${3}/video-playlists?start=${0}&count=${1}",
            key="data"
        )
