from quart import render_template
from datetime import datetime
from dateutil import parser as dateutil
import peertube
import html2text
import sys

# Extra information about video, not contained directly in API result
# a is video result
class VideoInfo:
    def __init__(self, api, a, args):
        # If the video is being built from a failed request, return that request instead
        if isinstance(a, FailedRequest):
            print("[ERROR] A video request failed, yet you called VideoInfo about it. You should probably not make useless requests.")
            return FailedRequest("A previous video request failed, not attempting to fetch extra information about it.")

        quality = args.get("quality")

        self.resolutions = []
        self.video = None

        self.files = a["files"]
        if len(self.files) == 0:
            self.files = ((a["streamingPlaylists"])[0])["files"]

        self.default_res = None

        for entry in self.files:
            resolution = (entry["resolution"])["id"]
            self.resolutions.append(entry["resolution"])

            # chose the default quality
            if resolution != 0 and quality == None:
                if self.default_res == None:
                    self.default_res = resolution
                    self.video = entry["fileUrl"]
                elif abs(720 - resolution) < abs(720 - self.default_res):
                    self.default_res = resolution
                    self.video = entry["fileUrl"]

            if str(resolution) == str(quality):
                self.video = entry["fileUrl"]

        if quality == None:
            self.quality = self.default_res
        else:
            self.quality = quality

        self.no_quality_selected = not self.video

# --- IDENTIFIERS UTILS ---
# Builds a unified id@server from one of those syntaxes, additionally stripping extra whitespace and ignoring `#` as comments:
#   - id@server
#   - @id@server
#   - http(s)://server/c/id
#   - http(s)://server/a/id
def find_subscription(request):
    identifier = request
    identifier = identifier.split('#')[0].strip()
    # Comment line is returned as empty string
    if identifier == '': return ''
    if identifier.startswith('@'):
        # Strip @ from identifier
        return identifier[1:]

    if identifier.startswith('http'):
        identifier = identifier[4:]
        # HTTPS?
        if identifier.startswith('s'): identifier = identifier[1:]
        # Remove ://
        identifier = identifier[3:]
        parts = identifier.split('/')
        domain = parts[0]
        if parts[1] == 'a' or parts[1] == 'c':
            # Account or channel found, take the next part
            return parts[2] + '@' + domain
    else:
        # Just check there's an @ in there and it should be fine
        if '@' in identifier:
            return identifier

    # No match was found, we don't understand this URL
    print("[WARN] Identifier not understood from local subscriptions:\n%s" % request)
    return ''

def account_name_domain(account):
    name, domain = account.split('@')
    return name, domain

def account_domain_name(account):
    name, domain = account.split('@')
    return domain, name
# --- IDENTIFIERS HELPERS ---

# --- TEMPLATING HELPERS ---
# Feed me a template name, a peertube.Api.method() result,
# and a lamba to build the template context from result, on success
# If passed result is in fact an iterable (list/tuple) of results,
# they will all be checked for success
async def render(template, result, context):
    if hasattr(result, '__iter__'):
        failures = []
        for res in result:
            if isinstance(res, FailedRequest):
                failures.append(res.format_html())
        if failures:
            # One or more failures occurred
            return await render_template(
                "error.html",
                error_number = "500",
                error_reasons = failures,
            ), 500
    elif isinstance(result, FailedRequest):
        return await render_template(
            "error.html",
            error_number = "500",
            error_reasons = [result.format_html()],
        ), 500
    context = context(result)
    try:
        return await render_template(
            template,
            **context,
        )
    except Exception as e:
        return await render_error(FailedRequest(e))

# Feed me a single peertube.Api that failed (FailedRequest class), i will render an error
async def render_error(result):
    if not isinstance(result, peertube.FailedRequest):
        print("[ERROR] Called render_error with a non-failed result:\n%s" % result)
        return
    return await render_template(
        "error.html",
        error_number = "500",
        error_reasons = [result.format_html()],
    ), 500

# Feed me an error code and a reason, i will render an error
async def render_custom_error(reason, code=500):
    return await render_template(
        "error.html",
        error_number = str(code),
        error_reasons = [reason],
    ), code

# Feed me a peertube.Api result and i will output plaintext, or render an error
async def render_plaintext(result):
    if isinstance(result, peertube.FailedRequest):
        return await render_template(
            "error.html",
            error_number = "500",
            error_reasons = [result.format_html()],
        ), 500
    return result
# --- END TEMPLATING UTILS ---

# --- COMMENT UTILS ---
h2t = html2text.HTML2Text()
h2t.ignore_links = True

def format_comments(comments):
    if isinstance(comments, FailedRequest):
        return comments
    # Strip the HTML from the comments and convert them to plain text
    new_comments = []
    for comment in comments:
        text = h2t.handle(comment["text"]).strip().strip("\n")
        comment["text"] = text
        new_comments.append(comment)
    return new_comments
# --- END COMMENT UTILS ---

# --- SUBSCRIPTION UTILS ---
# Manage Subscriptions to remote channels and accounts (see INDEX ROUTE in main.py)
class Subscriptions:
    def __init__(self, folder, api, cache, ttl=3600):
        self.api = api
        self.cache = cache
        self.folder = folder
        self.ttl = ttl

    # Returns an entry from cache, automatically prefixing "peertube-FUNC-" where FUNC is the caller method name
    def cached(self, key, ttl=None):
        if ttl == None: ttl = self.ttl
        caller = sys._getframe(1).f_code.co_name
        key_name = "subscriptions-" + caller + "-" + key
        res = self.cache.get(key_name, ttl=ttl)
        if res == None:
            print("[CACHE] Entry not found for %s: %s" % (caller, key))
            return None
        print("[CACHE] Found entry for %s: %s" % (caller, key))
        return res

    # Save an entry into cache, automatically prefixing "peertube-FUNC-" where FUNC is the caller method name
    def save(self, key, value):
        caller = sys._getframe(1).f_code.co_name
        key_name = "subscriptions-" + caller + "-" + key
        print("[CACHE] Saving entry for %s: %s" % (caller, key))
        self.cache.set(key_name, value)

    # Load subscriptions from a file called `kind`.list (60s cache)
    def load(self, kind):
        print("[INFO] Refreshing subscriptions %s from %s/%s.list" % (kind, self.folder, kind))
        try:
            with open(self.folder + '/' + kind + '.list', 'r') as f:
                subscriptions = map(find_subscription, f.read().splitlines())
        except Exception as e:
            print("[INFO] No `%s.list` file to load for local subscriptions in %s, or the file isn't readable" % (kind, self.folder))
            subscriptions = []
        # Remove comment entries and empty lines
        results = list(filter(lambda entry: entry != '', subscriptions))
        print("[DEBUG] %s subscriptions: %s" % (kind, results))
        return results

    # Get local subscriptions for accounts and channels
    def list(self, ttl=None):
        list = []
        list.extend(self.accounts(ttl=ttl))
        list.extend(self.channels(ttl=ttl))
        return list

    # Get the info about local subscriptions for accounts and channels. Returns a list of successes
    # and a list of failures
    def info(self, ttl=None):
        list = []
        failures = []

        accounts_info, accounts_failures = self.accounts_info(ttl=ttl)
        list.extend(accounts_info)
        failures.extend(accounts_failures)
        
        channels_info, channels_failures = self.channels_info(ttl=ttl)
        list.extend(channels_info)
        failures.extend(channels_failures)

        return list, failures

    # Get the latest `limit` videos from accounts and channels subscriptions combined. Returns a list of successes and failures
    # NOTE: duplicates are not handled, why would you add both an account and the corresponding channel?
    def videos(self, limit=12, ttl=None):
        latest = []
        failures = []
        
        accounts_latest, accounts_failures = self.accounts_videos(limit=limit, ttl=ttl)
        latest.extend(accounts_latest)
        failures.extend(accounts_failures)

        channels_latest, channels_failures = self.channels_videos(limit=limit, ttl=ttl)
        latest.extend(channels_latest)
        failures.extend(channels_failures)
        
        # TODO: maybe refactor so we don't have to reorder twice?
        latest.sort(key = lambda vid: dateutil.isoparse(vid["createdAt"]), reverse=True)
        return latest[0:limit], failures

    # List of locally-subscribed accounts (accounts.list)
    # Enforce 60s refresh by default
    def accounts(self, ttl=60):
        cached = self.cached("accounts", ttl=ttl)
        if cached == None:
            # Refresh local accounts subscriptions
            res = self.load("accounts")
            self.save("accounts", res)
            return res
        return cached

    # List of detailed info about local account subscriptions
    def accounts_info(self, ttl=None):
        # Both are cached, if we cache again we risk creating even more stale data
        info = []
        failures = []
        for sub in self.accounts(ttl=ttl):
            result = self.api.account(*account_domain_name(sub))
            if isinstance(result, FailedRequest):
                # Single failure from one API request
                failures.append(result)
            else:
                # Single success from one API request
                info.append(result)
        return info, failures

    # List of latest videos from account subscriptions
    # Some requests are bound to fail, so we return a tuple of (results[], failures[])
    def accounts_videos(self, limit=12, ttl=None):
        latest  = []
        failures = []
        for sub in self.accounts(ttl=ttl):
            result = self.api.account_videos(*account_domain_name(sub))
            if isinstance(result, FailedRequest):
                # Single failure from one API request
                failures.append(result)
            else:
                # Multiple successes from one API request
                latest.extend(result)
        latest.sort(key = lambda vid: dateutil.isoparse(vid["createdAt"]), reverse=True)
        return (latest[0:limit], failures)

    # List of locally-subscribed channels (channels.list)
    # Enforce 60s refresh by default
    def channels(self, ttl=60):
        cached = self.cached("channels", ttl=ttl)
        if cached == None:
            # Refresh local channels subscriptions
            res = self.load("channels")
            self.save("channels", res)
            return res
        return cached

    # List of detailed info about local channel subscriptions
    # Returns a list of successes, and a list of failures
    def channels_info(self, ttl=None):
        # Both are cached, if we cache again we risk creating even more stale data
        info = []
        failures = []
        for sub in self.channels(ttl=ttl):
            result = self.api.channel(*account_domain_name(sub))
            if isinstance(result, FailedRequest):
                # Single failure from one API request
                failures.append(result)
            else:
                # Single success from one API request
                info.append(result)
        return info, failures

    # List of latest videos from channel subscriptions
    # Some requests are bound to fail, so we return a tuple of (results[], failures[])
    def channels_videos(self, limit=12, ttl=None):
        latest  = []
        failures = []
        for sub in self.channels(ttl=ttl):
            result = self.api.channel_videos(*account_domain_name(sub))
            if isinstance(result, FailedRequest):
                # Single failure from one API request
                failures.append(result)
            else:
                # Multiple successes from one API request
                latest.extend(result)

        latest.sort(key = lambda vid: dateutil.isoparse(vid["createdAt"]), reverse=True)
        return latest[0:limit], failures
# --- END SUBSCRIPTION UTILS ---

# --- ERROR UTILS ---
# Semantic class to store remote errors
class FailedRequest:
    def __init__(self, e):
        self.exception = e

    def format(self):
        # If it's a rawtext error, print it
        # Otherwise look for 'message' attribute
        # Otherwise ask python to represent the exception
        if isinstance(self.exception, str):
            return self.exception
        else: return getattr(self.exception, 'message', repr(self.exception))

    # Strip < and > symbols to prevent content injection,
    # and replace newlines with HTML line breaks <br>
    def format_html(self):
        return self.format().replace('<', "&lt;").replace('>', "&gt;").replace("\n", "<br>")

# Format a list of FailedRequest's
def format_errors(failures):
    return list(map(lambda failure: failure.format(), failures))

def format_errors_html(failures):
    return list(map(lambda failure: failure.format_html(), failures))
# --- END ERROR UTILS ---
